from __future__ import annotations

import abc
import functools
from typing import Any, Optional, Protocol, Union, Sequence, ClassVar, Type

from np_workflows.services.protocols import (
    Service,
    TestFailure,
    Configurable,
    Initializable,
    Testable,
    Pretestable,
    Startable,
    Verifiable,
    Stoppable,
    Finalizable,
    Validatable,
    Shutdownable,
)
from np_workflows.services import config

import np_session
import np_logging

logger = np_logging.getLogger(__name__)

class Experiment(abc.ABC):
    
    rig_idx: ClassVar[int] = config.Rig.idx
    # session: str
    # "Unique id for the session, e.g. lims ecephys session id, datetime. Used for folder name."
    services: tuple[Service, ...]
    "Devices, databases etc. Will always be called in order."
    
    config: dict[str, dict[str, Any]]
    "Top-level keys are names of Services. Each Service then has config specific to rig, session, experiment, etc."
    
    state_globals: dict[str, dict[str, Any]]
    "Storage for WSE state dictionary. Main key of interest is `external`"
    
    def apply_config_to_services(self) -> None:
        """For each service, apply every key in self.config['service'] as an attribute."""        
        def apply_config(service):
            if config := self.config['services'].get(service.__name__):
                for key, value in config.items():
                    setattr(service, key, value)
                    logger.debug(f"{self.__class__.__name__} | Set {service.__name__}.{key} = {getattr(service, key)}")
        
        for service in self.services:
            for base in service.__class__.__bases__:
                apply_config(base)
            apply_config(service)
            
    # def configure_services(self) -> None:
    #     for service in (_ for _ in self.services if isinstance(_, Configurable)):
    #         service.configure()
            
    def initialize_services(self) -> None:
        
        self.apply_config_to_services()
        self.state_globals['external']['component_status'] = {service.__name__: False for service in self.services}
        for service in self.services:
                        
            if isinstance(service, Initializable):
                while True:
                    try:
                        service.initialize()
                        
                    except Exception as exc:
                        logger.error("%s | %r", service.__name__, exc)
                        import pdb; pdb.set_trace()
                        continue
                    
                    else:
                        break
            
            if isinstance(service, Testable):
                while True:
                    try:
                        service.test()
                        
                    except TestFailure as exc:
                        try:
                            logger.error("%s | %r", service.__name__, service.exc)
                        except AttributeError:
                            logger.error("%s | %r", service.__name__, exc)
                        import pdb; pdb.set_trace()
                        continue
                    
                    except Exception as exc:
                        logger.error("%s | %r", service.__name__, exc)
                        import pdb; pdb.set_trace()
                        continue
                    
                    else:
                        break
            
            self.state_globals['external']['component_status'][service.__name__] = True
            
                
    def pretest_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Pretestable)):
            service.pretest()
            
    def start_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Startable)):
            service.start()
            if isinstance(service, Verifiable):
                service.verify()
                
    def stop_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Stoppable)):
            service.stop()
            if isinstance(service, Finalizable):
                service.finalize()
                
    def validate_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Validatable)):
            service.validate()
            
    def shutdown_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Shutdownable)):
            service.shutdown()
    
    
class WithLims(Experiment):
    "Provides lims info properties"
    
    def __init__(self,
        labtracks_mouse_id: int | str,
        lims_user_id: str,
        lims_session_id: Optional[int] = None,
    ):
        logger.debug(f"Initializing {__class__.__name__}({labtracks_mouse_id}, {lims_user_id})")
        self.mouse = labtracks_mouse_id
        self.operator = lims_user_id
        if lims_session_id is None:
            try:
                lims_session_id = self.generate_ecephys_session()
            except Exception as exc:
                logger.exception(exc)
                import pdb; pdb.set_trace()
        self.session = lims_session_id
        logger.debug(f"{__class__.__name__}.mouse = {self.mouse!r}")
        logger.debug(f"{__class__.__name__}.operator = {self.operator!r}")
        logger.debug(f"{__class__.__name__}.session = {self.session!r}")
        logger.debug(f"{__class__.__name__}.folder = {self.folder}")
        

    def generate_ecephys_session(
        self,
        labtracks_mouse_id: Optional[int | str] = None,
        lims_user_id: Optional[str] = None,
    ) -> np_session.lims.SessionInfo:
        if labtracks_mouse_id is None:
            labtracks_mouse_id = self.mouse
        if lims_user_id is None:
            lims_user_id = self.operator
        return np_session.lims.generate_ecephys_session(mouse=labtracks_mouse_id, user=lims_user_id)

    @property
    def session(self) -> np_session.lims.SessionInfo:
        if not hasattr(self, '_session'):
            self._session = self.generate_ecephys_session()
        return self._session

    @session.setter
    def session(self, value: str | np_session.lims.SessionInfo):
        if not isinstance(value, np_session.lims.SessionInfo):
            value = np_session.lims.SessionInfo(value)
        self._session = value

    @property
    def mouse(self) -> np_session.lims.MouseInfo:
        return self._mouse

    @mouse.setter
    def mouse(self, value: str | int):
        self._mouse = np_session.lims.MouseInfo(value)

    @property
    def operator(self) -> np_session.lims.UserInfo:
        return self._operator

    @operator.setter
    def operator(self, value: str):
        self._operator = np_session.lims.UserInfo(value)

    @property
    def project(self) -> str:
        "Primarily used for file manifest"
        if not hasattr(self, '_project'):
            self._project = self.mouse.project
        return self._project

    @project.setter
    def project(self, value: str):
        self._project = value

    @functools.cached_property
    def folder(self) -> str:
        return np_session.folder_from_lims_id(str(self.session))

    @functools.cached_property
    def files(self) -> dict[str, dict[str, str]]:
        "Expected manifest of files from experiment"
        return np_session.files_manifest(str(self.project), self.folder, 'D1')




from __future__ import annotations
import datetime
import time 

try:
    #! the wse allows import errors to pass silently!
    #* put all imports in this try block so that we can see the error before exiting
    
    import inspect
    from typing import Any, Type

    from np_workflows.experiments import baseclasses, classes
    from np_workflows.experiments.baseclasses import Experiment
    from np_workflows.services.protocols import Initializable, Testable, TestFailure
    from np_workflows.services.config import Rig
    from np_workflows.services import zro, utils
    
    import np_session
    import np_logging
    import np_config
    
except Exception as exc:
    print(repr(exc))
    import pdb; pdb.set_trace()
    quit()
    
logger = np_logging.getLogger(__name__)

# Assign default values to global variables so they can be imported elsewhere
experiment: Experiment | str = ''
operator: np_session.lims.UserInfo | str = ''
mouse: np_session.lims.MouseInfo | str = ''
is_mouse_in_lims: bool | None = None
session: np_session.lims.SessionInfo | str | int = ''

# For reference:
transition_result: bool = True # T/F just sets text at bottom green/red, then returns to green in proceeding state
status_message: str = 'Ok'
msg_text: str = ''
alert: bool = False # T adds warning icon


def current_state() -> tuple[str, str]:
    "State name (before last underscore) and transition type (enter, input, exit)"
    current_frame = inspect.currentframe()
    calling_frame = inspect.getouterframes(current_frame, 2)[1][3] #?
    return calling_frame[: calling_frame.rfind("_")], calling_frame[calling_frame.rfind("_") + 1 :]

experiments: tuple[Type[Experiment], ...] = (
    classes.Pretest,
    classes.NpUltra,
    ) # TODO plug-in experiments

# TODO get userlist from np_config on per-rig basis
lims_user_ids: tuple[str, ...] = (         
        "hannah.belski",
        "hannah.cabasco",
        "ryan.gillis",
        "henry.loeffler",
        "corbettb",
        "ben.hardcastle",
        "samg",
        "ethan.mcbride",
        "jackie.kuyat",
        "andrew.shelton",
    )

def get_operators() -> list[str]:
    return list(lims_user_ids)
    
def get_experiments(with_lims: bool = True) -> list[str]:
    return [cls.__name__ for cls in experiments if with_lims or not issubclass(cls, baseclasses.WithLims)]

def start_rsc_apps() -> None:
    rsc_app_ids_required = np_config.from_zk(f'/rigs/{Rig.ID}')['rsc_app_ids_required']
    for comp in (Rig.Sync, Rig.Stim, Rig.Mon, Rig.Acq):
        rsc_node = zro.Proxy(comp.host, 6000)
        status = rsc_node.p_status()
        if any(apps_required_in_node := [_ for _ in rsc_app_ids_required if _ in status]):
            for app in apps_required_in_node:
                rsc_node.p_start(app)
        
def set(state_globals, **kwargs) -> None:
    "Update `state_globals['external'][key]` with kwargs"
    state_globals['external'].update(kwargs)

def get(state_globals, *args: str) -> Any | None | tuple[Any, ...]:
    "Get `state_globals['external'][arg]` for each arg in args"
    results = [state_globals['external'].get(arg) for arg in args]
    if len(results) == 1:
        return results[0]
    return tuple(results) if any(results) else None

def initialize_services(state_globals) -> None:
    experiment.apply_config_to_services()
    for service in experiment.services:
        
        set(state_globals, component_status={service.__name__: False})
        
        if isinstance(service, Initializable):
            while True:
                try:
                    service.initialize()
                    
                except Exception as exc:
                    logger.error("%s | %r", service.__name__, exc)
                    import pdb; pdb.set_trace()
                    continue
                
                else:
                    break
        
        if isinstance(service, Testable):
            while True:
                try:
                    service.test()
                    
                except TestFailure as exc:
                    try:
                        logger.error("%s | %r", service.__name__, service.exc)
                    except AttributeError:
                        logger.error("%s | %r", service.__name__, exc)
                    import pdb; pdb.set_trace()
                    continue
                
                except Exception as exc:
                    logger.error("%s | %r", service.__name__, exc)
                    import pdb; pdb.set_trace()
                    continue
                
                else:
                    break
                
def await_timer(state_globals: dict, wait_sec: int) -> datetime.timedelta:
    """
    Return timedelta until wait_sec is reached, then returns timedelta(0) which
    evaluates to False in Boolean contexts.
    
    Usage::
        >>> while time_remaining := await_timer(state_globals, wait_sec=600):
                print(f'{time_remaining} remaining')
                time.sleep(.1)
            print('Finished waiting')
    """
    if not get(state_globals, 'timer_start'):
        set(state_globals, timer_start=time.time()) # can't put datetime obj in state_globals['external']
    
    time_remaining = datetime.timedelta(
        seconds = wait_sec - (time.time() - get(state_globals, 'timer_start'))
        )
    
    if time_remaining.total_seconds() >= 0:
        print(f'{time_remaining} remaining')
        return time_remaining
    set(state_globals, timer_start=0) # reset for re-use
    return datetime.timedelta(0)


from __future__ import annotations

try:
    #! the wse allows import errors to pass silently!
    #* put all imports in this try block so that we can see the error before exiting
    
    import datetime
    import time
    from typing import Type

    from np_workflows.experiments import baseclasses, classes
    from np_workflows.experiments.baseclasses import Experiment
    from np_workflows.workflows.shared import npxc

    # we need all experiment classes imported into this namespace
    from np_workflows.experiments.classes import *

    import np_session
    import np_logging

except Exception as exc:
    print(repr(exc))
    import pdb; pdb.set_trace()
    quit()
    
logger = np_logging.getLogger(__name__)


# mouse & operator --------------------------------------------------------------------- 

def select_mouse_and_operator_enter(state_globals) -> None:
    npxc.set(state_globals, operators=[""] + npxc.get_operators())
    state_globals["external"]["mouse"] = '366122' #! defaults for testing
    # state_globals["external"]["operator"] = 'ben.hardcastle' #! defaults for testing
    
    # import pdb; pdb.set_trace()
    # state_globals['resources']['workflow']['states'].keys()
     
def select_mouse_and_operator_input(state_globals) -> None:
        
    mouse_input = state_globals["external"]["mouse"]
    try:
        npxc.mouse = np_session.lims.MouseInfo(mouse_input)
    except KeyError as exc:
        logger.warning("LIMS | Mouse %r could not be fetched from database: %r", mouse_input, exc)
        npxc.mouse = mouse_input
        npxc.is_mouse_in_lims = False
    
    try:
        project = npxc.mouse.project_name
    except (AttributeError, KeyError):
        project = "NA"
        if npxc.is_mouse_in_lims is None:
            logger.warning("LIMS | Mouse %r could not be fetched from database: %r", mouse_input)
        npxc.mouse = mouse_input
        npxc.is_mouse_in_lims = False
    else:  
        npxc.is_mouse_in_lims = True
    finally:
        npxc.set(state_globals, project=str(project), is_mouse_in_lims=npxc.is_mouse_in_lims)
    
    operator_input = state_globals["external"]["operator"]
    try:
        npxc.operator = np_session.lims.UserInfo(operator_input)
    except KeyError as exc:
        logger.warning("LIMS | Operator %r could not be fetched from database: %r", operator_input, exc)
        npxc.operator = operator_input
        
    status_message = f'Mouse {"is" if npxc.is_mouse_in_lims else "is not"} in lims'
    transition_result = False if not npxc.is_mouse_in_lims else True # None isn't allowed in s_g['external'] 
    npxc.set(state_globals, status_message=status_message, transition_result=transition_result)
        
# experiment ---------------------------------------------------------------------------

def select_experiment_enter(state_globals) -> None:
    experiments = npxc.get_experiments(with_lims = bool(npxc.is_mouse_in_lims))
    if not experiments:
        alert = True
        msg_text = "No experiments found"
        npxc.set(state_globals, alert=alert, msg_text=msg_text)
    npxc.set(state_globals, experiments=[""] + experiments)
    
    
def select_experiment_input(state_globals) -> None:
    "Create Experiment instance and configure services"
    
    experiment_class: Type[Experiment] = eval(npxc.get(state_globals, 'experiment'))
    
    if issubclass(experiment_class, baseclasses.WithLims):
        npxc.experiment = experiment_class(
            *npxc.get(state_globals, 'mouse', 'operator'),
            lims_session_id = int(npxc.session) if npxc.session else None,
            )
    else:
        npxc.experiment = experiment_class(npxc.mouse)
        
    if not isinstance(npxc.experiment, Experiment):
        raise ValueError(f"Experiment {npxc.experiment!r} was not converted to a valid Experiment class")
        
    npxc.experiment.state_globals = state_globals
    npxc.experiment.initialize_services()

    
